/* 
 * scannedonly VFS module for Samba 3.0, 3.2 and 3.4
 *
 * Copyright 2007,2008,2009 (C) Olivier Sessink
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ABOUT SCANNEDONLY
 *
 * scannedonly implements a 'filter' like vfs module that talks over a
 * unix domain socket or over UDP to a anti-virus engine.
 *
 * files that are clean have a corresponding .scanned:{filename} file
 * in the same directory. So why the .scanned: files? They take up 
 * only an inode, because they are 0 bytes. To test if the file is 
 * scanned only a stat() call on the filesystem is needed which is 
 * very quick compared to a database lookup. All modern filesystems 
 * use database technology such as balanced trees for lookups anyway. 
 * The number of inodes in modern filesystems is also not limiting 
 * anymore. The .scanned: files are also easy scriptable. You can 
 * remove them with a simple find command or create them with a 
 * simple touch command. Extended filesystem attributes have similar
 * properties, but are not supported on all filesystems, so that
 * would limit the usage of the module (and attributes are not as 
 * easily scriptable)
 *
 * files that are not clean are sent to the AV-engine. Only the 
 * filename is sent over the socket. The protocol is very simple:
 * a newline separated list of filenames inside each datagram.
 *
 * a file AV-scan may be requested multiple times, the AV-engine 
 * should also check if the file has been scanned already. Requests
 * can also be dropped by the AV-engine (and we thus don't need the
 * reliability of TCP).
 * 
 */

#include "includes.h"

#include "config.h"
struct Tscannedonly {
	int socket;
	int domain_socket;
	char *socketname;
	int portnum;
	char *scanhost;
	int show_special_files;
	int rm_hidden_files_on_rmdir;
	int hide_nonscanned_files;
	int allow_nonscanned_files;
	char *scanning_message;
	int scanning_message_len;
	int recheck_time_open;
	int recheck_tries_open;
	int recheck_size_open;
	int recheck_time_readdir;
	int recheck_tries_readdir;
	char *activefiles[1024];
};

struct scannedonly_DIR {
	char *basename;
	int notify_loop_done;
	SMB_STRUCT_DIR *DIR;
};

/*********************/
/* utility functions */
/*********************/

static char *real_path_from_notify_path(struct Tscannedonly *so,
					const char *path)
{
	char *name;
	int len, pathlen;

	name = strrchr(path, '/');
	if (!name)
		return NULL;
	pathlen = name - path;
	name++;
	len = strlen(name);
	if (len <= so->scanning_message_len)
		return NULL;

	if (strcmp
	    (name + (len - so->scanning_message_len),
	     so->scanning_message) != 0)
		return NULL;

	return strndup(path, pathlen + len - so->scanning_message_len);
}

static char *cachefile_name(const char *shortname, const char *basename)
{
	char *cachefile;
	int rlen = strlen(shortname);
	int blen = strlen(basename);
	DEBUG(9,
	      ("cachefile_name shortname=%s basename=%s\n", shortname,
	       basename));
	cachefile = (char *)SMB_MALLOC(rlen + blen + 9 + 1);
	cachefile =
	    strncat(strncat
		    (strncpy(cachefile, basename, blen + 1), ".scanned:",
		     9 + 1), shortname, rlen + 1);
	return cachefile;
}

static char *name_w_ending_slash(const char *name)
{
	int len;

	len = strlen(name);
	if (name[len - 1] == '/') {
		return SMB_STRDUP(name);
	} else {
		char *tmp;
		tmp = SMB_MALLOC(len + 2);
		strncpy(tmp, name, len + 1);
		return strncat(tmp, "/", 1);
	}
}

static char *cachefile_name_f_fullpath(const char *fullpath)
{
	char *tmp, *cachefile, *shortname, *basename;
	int free_basename = 0;
	tmp = strrchr(fullpath, '/');
	if (tmp) {
		basename = SMB_STRNDUP(fullpath, (tmp - fullpath) + 1);
		free_basename = 1;
		shortname = tmp + 1;
	} else {
		basename = "";
		shortname = (char *)fullpath;
	}
	cachefile = cachefile_name(shortname, basename);
	if (free_basename) {
		SAFE_FREE(basename);
	}
	return cachefile;
}

static char *path_plus_name(const char *basename, const char *filename)
{
	int baselen = strlen(basename);
	int filelen = strlen(filename);

	return
	    strncat(strncpy
		    (SMB_MALLOC(baselen + filelen + 1), basename,
		     baselen + 1), filename, filelen + 1);
}

static char *construct_full_path(vfs_handle_struct * handle,
				 const char *somepath, int ending_slash)
{
	if (!somepath)
		return NULL;
	if (somepath[0] == '/') {
		if (ending_slash)
			return name_w_ending_slash(somepath);
		else
			return SMB_STRDUP(somepath);
	} else {
		char *fullpath, *tmp;
		const char *iter;
		int tmplen, rellen, itlen;
#ifdef SAMBA_3_0
		TALLOC_CTX *ctx = NULL;
#else
		TALLOC_CTX *ctx = talloc_tos();
#endif
		tmp = vfs_GetWd(ctx, handle->conn);
		/*getcwd(tmp, 1024); */
		DEBUG(9,
		      ("construct_full_path, pwd=%s, relpath=%s\n", tmp,
		       somepath));
		tmplen = strlen(tmp);
		rellen = strlen(somepath);
		fullpath = SMB_MALLOC(tmplen + rellen + 4);
		fullpath = strncpy(fullpath, tmp, tmplen + 1);
		if (fullpath[tmplen - 1] != '/') {
			fullpath = strncat(fullpath, "/", 2);
		}
		iter = somepath;
		while (*iter == '.' && *(iter + 1) == '/')
			iter += 2;
		fullpath = strncat(fullpath, iter, rellen + 1);
		/* iter can be empty */
		itlen = strlen(iter);
		if (ending_slash && itlen > 0 && iter[itlen - 1] != '/') {
			fullpath = strncat(fullpath, "/", 2);
		}
		talloc_free(tmp);
		return fullpath;
	}
}

static int connect_to_scanner(vfs_handle_struct * handle)
{
	struct Tscannedonly *so = (struct Tscannedonly *)handle->data;

	if (so->domain_socket) {
		struct sockaddr_un saun;
		DEBUG(9, ("socket=%s\n", so->socketname));
		if ((so->socket = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
			DEBUG(9, ("failed to create socket\n"));
			return -1;
		}
		saun.sun_family = AF_UNIX;
		strncpy(saun.sun_path, so->socketname,
			strlen(so->socketname) + 1);
		if (connect(so->socket, &saun, SUN_LEN(&saun)) < 0) {
			DEBUG(9, ("failed to connect to socket\n"));
			return -1;
		}
		DEBUG(9,
		      ("bound %s to socket %d\n", saun.sun_path, so->socket));

	} else {
		struct sockaddr_in sain;
		struct hostent *he;
		DEBUG(9, ("host=%s, port=%d\n", so->scanhost, so->portnum));
		if ((he = gethostbyname(so->scanhost)) == NULL) {
			DEBUG(2,
			      ("failed to resolve hostname %s\n",
			       so->scanhost));
			return -1;
		}
		if ((so->socket = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			DEBUG(2,
			      ("failed to create socket to %s\n",
			       so->scanhost));
			return -1;
		}
		memset(&sain, 0, sizeof(sain));
		sain.sin_family = AF_INET;
		sain.sin_port = htons(so->portnum);
		sain.sin_addr = *((struct in_addr *)he->h_addr);

		if (connect(so->socket, &sain, sizeof(sain)) < 0) {
			DEBUG(2,
			      ("failed to connect to socket to %s\n",
			       so->scanhost));
			return -1;
		}
		DEBUG(4,
		      ("bound port %d to socket %d\n", so->portnum,
		       so->socket));
	}

	{
		int sndsize;
		socklen_t size = sizeof(int);
		getsockopt(so->socket, SOL_SOCKET, SO_RCVBUF,
			   (char *)&sndsize, &size);
		DEBUG(9, ("current socket buffer size=%d\n", sndsize));
		sndsize = 262144;
		if (setsockopt
		    (so->socket, SOL_SOCKET, SO_RCVBUF, (char *)&sndsize,
		     (int)sizeof(sndsize)) != 0) {
			DEBUG(9,
			      ("error setting socket buffer %s (%d)\n",
			       strerror(errno), errno));
		}
	}

	if (fcntl(so->socket, F_SETFL, O_NDELAY) < 0) {
		DEBUG(9,
		      ("error setting socket NONBLOCKING %s (%d)\n",
		       strerror(errno), errno));
	}

	return 0;
}

#define SENDBUFFERSIZE 1450
static char gsendbuffer[SENDBUFFERSIZE + 1];

static void flush_sendbuffer(vfs_handle_struct * handle)
{
	struct Tscannedonly *so = (struct Tscannedonly *)handle->data;
	int ret, len, loop = 10;
	if (gsendbuffer[0] == '\0')
		return;

	do {
		loop--;
		len = strlen(gsendbuffer);
		ret = send(so->socket, gsendbuffer, len, MSG_DONTWAIT);
		if (ret == len) {
			gsendbuffer[0] = '\0';
			break;
		}
		if (ret == -1) {
			DEBUG(3,("scannedonly flush_sendbuffer: "
			    "error sending on socket %d to scanner: %s (%d)\n",
			    so->socket, strerror(errno), errno));
			if (errno == ECONNREFUSED || errno == ENOTCONN
			    || errno == ECONNRESET) {
				if (connect_to_scanner(handle) == -1)
					break;	/* connecting fails, abort */
				/* try again */
			} else if (errno != EINTR) {
				/* on EINTR we just try again, all remaining 
				other errors we log the error
				   and try again ONCE */
				loop = 1;
				DEBUG(3,("scannedonly flush_sendbuffer: "
				   "error sending data to scanner: %s (%d)\n",
				   strerror(errno), errno));
			}
		} else {
			/* --> partial write: Resend all filenames that were 
			not or not completely written. a partial filename 
			written means the filename will not arrive correctly,
			so resend it completely */
			int pos = 0;
			while (pos < len) {
				char *tmp = strchr(gsendbuffer + pos, '\n');
				if (tmp && tmp - gsendbuffer < ret)
					pos = tmp - gsendbuffer + 1;
				else
					break;
			}
			memmove(gsendbuffer, gsendbuffer + pos,
				SENDBUFFERSIZE - ret);
			/* now try again */
		}
	} while (loop > 0);

	if (gsendbuffer[0] != '\0') {
		DEBUG(2,
		      ("scannedonly flush_sendbuffer: "
		       "failed to send files to AV scanner, "
		       "discarding files."));
		gsendbuffer[0] = '\0';
	}
}

static void notify_scanner(vfs_handle_struct * handle, const char *scanfile)
{
	char *tmp;
	bool free_tmp = false;
	int tmplen, gsendlen;
	if (scanfile[0] != '/') {
		tmp = construct_full_path(handle, scanfile, 0);
		free_tmp = true;
	} else {
		tmp = (char *)scanfile;
	}
	tmplen = strlen(tmp);
	gsendlen = strlen(gsendbuffer);
	DEBUG(9,
	      ("scannedonly notify_scanner: tmp=%s, tmplen=%d, gsendlen=%d\n",
	       tmp, tmplen, gsendlen));
	if (gsendlen + tmplen >= SENDBUFFERSIZE) {
		flush_sendbuffer(handle);
	}
	strncat(gsendbuffer, tmp, tmplen);
	strncat(gsendbuffer, "\n", 1);
	if (free_tmp) {
		SAFE_FREE(tmp);
	}
}

static int is_scannedonly_file(const char *shortname)
{
	if (strncmp(shortname, ".scanned:", 9) == 0) {
		return 1;
	}
	if (strncmp(shortname, ".virus:", 7) == 0) {
		return 1;
	}
	if (strncmp(shortname, ".failed:", 8) == 0) {
		return 1;
	}
	return 0;
}

#ifdef SAMBA_3_4
#define MY_NEXT_READDIR(handle, dirp) SMB_VFS_NEXT_READDIR(handle, dirp, NULL)
#else
#define MY_NEXT_READDIR(handle, dirp) SMB_VFS_NEXT_READDIR(handle, dirp)
#endif

/*
vfs_handle_struct *handle the scannedonly handle
scannedonly_DIR * sDIR the scannedonly struct if called from _readdir() or NULL
fullpath is a full path starting from /
shortname is the filename without directory components
basename, is the directory without file name component
allow_nonexistant return TRUE if stat() on the requested file fails
recheck_time, the time in milliseconds to wait for the daemon to create a 
.scanned file 
recheck_tries, the number of tries to wait
recheck_size, size in Kb of files that should not be waited for
loop : boolean if we should try to loop over all files in the directory and 
send a notify to the scanner for all files that need scanning
*/
static int scannedonly_allow_access(vfs_handle_struct * handle,
				    struct scannedonly_DIR *sDIR,
				    const char *fullpath,
				    const char *shortname,
				    const char *basename,
				    int allow_nonexistant,
				    int recheck_time, int recheck_tries,
				    int recheck_size, int loop)
{
	SMB_STRUCT_STAT sbuf1, sbuf2;
	char *cachefile;
	int retval;

	if ((strcmp(shortname, ".") == 0)
	    || (strcmp(shortname, "..") == 0)) {
		return 1;
	}
	if (is_scannedonly_file(shortname)) {
		DEBUG(5,
		      ("scannedonly_allow_access, %s is a scannedonly file, "
		       "return 0\n", shortname));
		return 0;
	}

	retval = SMB_VFS_NEXT_STAT(handle, fullpath, &sbuf1);
	if (retval != 0) {
		/* failed to stat this file?!? --> hide it */
		return allow_nonexistant;
	}
	if (!S_ISREG(sbuf1.st_mode)) {
		return (((struct Tscannedonly *)handle->data)->
			show_special_files || S_ISDIR(sbuf1.st_mode));
	}
	if (sbuf1.st_size == 0) {
		return 1;	/* empty files cannot contain viruses ! */
	}
	cachefile = cachefile_name(shortname, basename);
	retval = SMB_VFS_NEXT_STAT(handle, cachefile, &sbuf2);

	if ((retval == 0) && (sbuf1.st_mtime <= sbuf2.st_mtime)) {
		SAFE_FREE(cachefile);
		return 1;
	}

	/* no cachefile or too old */
	if (retval == 0) {
		SMB_VFS_NEXT_UNLINK(handle, cachefile);
		retval = -1;
	}

	notify_scanner(handle, fullpath);

	int didloop = 0;
	if (loop && sDIR && !sDIR->notify_loop_done) {
		/* check the rest of the directory and notify the 
		   scanner if some file needs scanning */
		long offset;
		SMB_STRUCT_DIRENT *dire;

		offset = SMB_VFS_NEXT_TELLDIR(handle, sDIR->DIR);
		dire = MY_NEXT_READDIR(handle, sDIR->DIR);
		while (dire) {
			char *fpath2;
			fpath2 = path_plus_name(basename, dire->d_name);
			DEBUG(9,
			      ("scannedonly_allow_access in loop, "
			       "found %s\n", fpath2));
			scannedonly_allow_access(handle, NULL,
						 fpath2,
						 dire->d_name,
						 basename, 0, 0, 0, 0, 0);
			SAFE_FREE(fpath2);
			dire = MY_NEXT_READDIR(handle, sDIR->DIR);
		} while (dire) ;
		sDIR->notify_loop_done = 1;
		didloop = 1;
		SMB_VFS_NEXT_SEEKDIR(handle, sDIR->DIR, offset);
	}
	if (recheck_time > 0
	    && ((recheck_size > 0 && sbuf1.st_size < (1024 * recheck_size))
		|| didloop)) {
		int i = 0;
		flush_sendbuffer(handle);
		while (retval != 0	/*&& errno == ENOENT */
		       && i < recheck_tries) {
			struct timespec req = { 0, recheck_time * 10000 };
			DEBUG(5,
			      ("scannedonly_allow_access, wait (try=%d "
			       "(max %d), %d ms) for %s\n",
			       i, recheck_tries, recheck_time, cachefile));
			nanosleep(&req, NULL);
			retval = SMB_VFS_NEXT_STAT(handle, cachefile, &sbuf2);
			i++;
		}
	}
	/* still no cachefile, or still too old, return 0 */
	if (retval != 0 || sbuf1.st_mtime > sbuf2.st_mtime) {
		SAFE_FREE(cachefile);
		return 0;
	}
	SAFE_FREE(cachefile);
	return 1;
}

/*********************/
/* VFS functions     */
/*********************/

static SMB_STRUCT_DIR *scannedonly_opendir(vfs_handle_struct * handle,
					   const char *fname,
					   const char *mask, uint32 attr)
{
	SMB_STRUCT_DIR * dirp;
	struct scannedonly_DIR *sDIR;
	
	dirp = SMB_VFS_NEXT_OPENDIR(handle, fname, mask, attr);
	if (!dirp)
		return NULL;
	sDIR = TALLOC_P(NULL, struct scannedonly_DIR);
	DEBUG(9,("scannedonly_opendir, open %s, created sDIR %p\n", fname, sDIR));
	if (fname[0] != '/')
		sDIR->basename = construct_full_path(handle, fname, 1);
	else
		sDIR->basename = name_w_ending_slash(fname);
	sDIR->DIR = dirp;
	sDIR->notify_loop_done = 0;
	return (SMB_STRUCT_DIR *) sDIR;
}

static SMB_STRUCT_DIRENT *scannedonly_readdir(vfs_handle_struct * handle,
					      SMB_STRUCT_DIR * dirp)
{
	SMB_STRUCT_DIRENT *result;
	int allowed = 0;
	char *tmp;

	struct scannedonly_DIR *sDIR = (struct scannedonly_DIR *)dirp;

	if (!dirp)
		return NULL;	
	
	result = MY_NEXT_READDIR(handle, sDIR->DIR);
	
	if (!result)
		return NULL;
	DEBUG(9,("scannedonly_readdir, result=%s\n", result->d_name));
	if (is_scannedonly_file(result->d_name)) {
		DEBUG(4,
		      ("scannedonly_readdir, %s is a scannedonly file, "
		       "skip to next entry\n", result->d_name));
		return scannedonly_readdir(handle, dirp);
	}

	tmp = path_plus_name(sDIR->basename, result->d_name);

	/* even if we don't hide nonscanned files or we allow non scanned 
	   files we call allow_access because it will notify the daemon to 
	   scan these files */
	allowed =
	    scannedonly_allow_access(handle, sDIR, tmp,
				     result->d_name,
				     sDIR->basename, 0,
				     ((struct Tscannedonly *)
				     	handle->data)->hide_nonscanned_files
				     ? ((struct Tscannedonly *)
					handle->data)->recheck_time_readdir : 0,
				     ((struct Tscannedonly *)
				      handle->data)->recheck_tries_readdir, -1,
				     1);
	DEBUG(4,
	      ("scannedonly_readdir access to %s (%s) = %d\n", tmp,
	       result->d_name, allowed));
	SAFE_FREE(tmp);

	if (allowed)
		return result;

	if (!((struct Tscannedonly *)handle->data)->hide_nonscanned_files
	    || ((struct Tscannedonly *)handle->data)->allow_nonscanned_files)
		return result;

	char *notify_name;
	int namelen;
	SMB_STRUCT_DIRENT *newdirent;
	DEBUG(3,
	      ("scannedonly_readdir, readdir listing for %s not "
	       "allowed, notify user\n", result->d_name));
	asprintf(&notify_name, "%s %s",
		 result->d_name,
		 ((struct Tscannedonly *)handle->data)->scanning_message);
	namelen = strlen(notify_name);
#ifdef SAMBA_3_0
	newdirent = (SMB_STRUCT_DIRENT *)
	TALLOC_ARRAY(sDIR, char, sizeof(SMB_STRUCT_DIRENT)
		     + namelen + 1);
#else
	newdirent = (SMB_STRUCT_DIRENT *)
	TALLOC_ARRAY(talloc_tos(), char, sizeof(SMB_STRUCT_DIRENT)
		     + namelen + 1);
#endif
	if (!newdirent) {
		return NULL;
	}
	memcpy(newdirent, result, sizeof(SMB_STRUCT_DIRENT));
	memcpy(&newdirent->d_name, notify_name, namelen + 1);
	SAFE_FREE(notify_name);
	DEBUG(5,
	      ("scannedonly_readdir, return newdirent at %p with "
	       "notification %s\n", newdirent, newdirent->d_name));
	return newdirent;
}

static void scannedonly_seekdir(struct vfs_handle_struct *handle,
				SMB_STRUCT_DIR * dirp, long offset)
{
	struct scannedonly_DIR *sDIR = (struct scannedonly_DIR *)dirp;
	SMB_VFS_NEXT_SEEKDIR(handle, sDIR->DIR, offset);
}

static long scannedonly_telldir(struct vfs_handle_struct *handle,
				SMB_STRUCT_DIR * dirp)
{
	struct scannedonly_DIR *sDIR = (struct scannedonly_DIR *)dirp;
	return SMB_VFS_NEXT_TELLDIR(handle, sDIR->DIR);
}

static void scannedonly_rewinddir(struct vfs_handle_struct *handle,
				  SMB_STRUCT_DIR * dirp)
{
	struct scannedonly_DIR *sDIR = (struct scannedonly_DIR *)dirp;
	SMB_VFS_NEXT_REWINDDIR(handle, sDIR->DIR);
}

static int scannedonly_closedir(vfs_handle_struct * handle,
				SMB_STRUCT_DIR * dirp)
{
	int retval;
	struct scannedonly_DIR *sDIR = (struct scannedonly_DIR *)dirp;
	flush_sendbuffer(handle);
	retval = SMB_VFS_NEXT_CLOSEDIR(handle, sDIR->DIR);
	SAFE_FREE(sDIR->basename);
	TALLOC_FREE(sDIR);
	return retval;
}

static int scannedonly_stat(vfs_handle_struct * handle, const char *path,
			    SMB_STRUCT_STAT * sbuf)
{
	int ret;
	ret = SMB_VFS_NEXT_STAT(handle, path, sbuf);
	DEBUG(5, ("scannedonly_stat: %s returned %d\n", path, ret));
	if (ret != 0 && errno == ENOENT) {
		char *realpath;
		/* possibly this was a fake name (file is being scanned for 
		   viruses.txt): check for that and create the real name and 
		   stat the real name */
		realpath = real_path_from_notify_path(((struct Tscannedonly *)
						       handle->data), path);
		if (realpath) {
			ret = SMB_VFS_NEXT_STAT(handle, realpath, sbuf);
			DEBUG(5, ("_stat: %s returned %d\n", realpath, ret));
			SAFE_FREE(realpath);
		}
	}
	return ret;
}

static int scannedonly_lstat(vfs_handle_struct * handle, const char *path,
			     SMB_STRUCT_STAT * sbuf)
{
	int ret;
	ret = SMB_VFS_NEXT_LSTAT(handle, path, sbuf);
	DEBUG(5, ("scannedonly_lstat: %s returned %d\n", path, ret));
	if (ret != 0) {
		char *realpath;
		/* possibly this was a fake name (file is being scanned for 
		   viruses.txt): check that and create the real name and stat 
		   the real name */
		realpath = real_path_from_notify_path(((struct Tscannedonly *)
						       handle->data), path);
		if (realpath) {
			ret = SMB_VFS_NEXT_LSTAT(handle, realpath, sbuf);
			DEBUG(5,
			      ("scannedonly_lstat: %s returned %d\n",
			       realpath, ret));
			SAFE_FREE(realpath);
		}
	}
	return ret;
}

static int scannedonly_open(vfs_handle_struct * handle, const char *fname,
			    files_struct * fsp, int flags, mode_t mode)
{
	char *basename, *tmp, *shortname;
	int allowed, free_basename = 0, write_access = 0, fd;
	/* if open for writing ignore it */
	if ((flags & O_ACCMODE) == O_WRONLY) {
		fd = SMB_VFS_NEXT_OPEN(handle, fname, fsp, flags, mode);
		if (fd < 1024) {
			((struct Tscannedonly *)handle->data)->activefiles[fd] =
			    construct_full_path(handle, fname, 0);
		}
		return fd;
	}
	if ((flags & O_ACCMODE) == O_RDWR) {
		write_access = 1;
	}
	/* check if this file is scanned already */
	tmp = strrchr(fname, '/');
	if (tmp) {
		basename = SMB_STRNDUP(fname, (tmp - fname) + 1);
		shortname = tmp + 1;
		free_basename = 1;
	} else {
		basename = "";
		shortname = (char *)fname;
	}
	allowed =
	    scannedonly_allow_access(handle, NULL, fname, shortname,
				     basename, write_access,
				     ((struct Tscannedonly *)handle->
				      data)->recheck_time_open,
				     ((struct Tscannedonly *)handle->
				      data)->recheck_tries_open,
				     ((struct Tscannedonly *)handle->
				      data)->recheck_size_open, 0);
	flush_sendbuffer(handle);
	if (free_basename)
		SAFE_FREE(basename);
	DEBUG(4, ("scannedonly_open: allow=%d for %s\n", allowed, fname));
	if (allowed
	    || ((struct Tscannedonly *)handle->data)->allow_nonscanned_files) {
		fd = SMB_VFS_NEXT_OPEN(handle, fname, fsp, flags, mode);
		if (write_access && fd < 1024) {
			((struct Tscannedonly *)handle->data)->activefiles[fd] =
			    construct_full_path(handle, fname, 0);
		}
		return fd;
	} else {
		errno = EACCES;
		return -1;
	}
}

#ifdef SAMBA_3_0
static void scannedonly_close(vfs_handle_struct * handle,
			      files_struct * fsp, int fd)
{
	if (fd < 1024
	    && ((struct Tscannedonly *)handle->data)->activefiles[fd] != NULL) {
		DEBUG(5,
		      ("scannedonly_close: request scanning for %s\n",
		       ((struct Tscannedonly *)handle->data)->activefiles[fd]));
		notify_scanner(handle,
			       ((struct Tscannedonly *)handle->
				data)->activefiles[fd]);
		flush_sendbuffer(handle);
		SAFE_FREE(((struct Tscannedonly *)handle->data)->
			  activefiles[fd]);
		((struct Tscannedonly *)handle->data)->activefiles[fd] = NULL;
	}
	SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
}
#else
static void scannedonly_close(vfs_handle_struct * handle, files_struct * fsp)
{
	/* retrieve the fd from fsp */

	if (fsp->fh->fd < 1024
	    && ((struct Tscannedonly *)handle->data)->activefiles[fsp->fh->fd]
	    != NULL) {
		DEBUG(5,
		      ("scannedonly_close: request scanning for %s\n",
		       ((struct Tscannedonly *)handle->data)->
		       activefiles[fsp->fh->fd]));
		notify_scanner(handle,
			       ((struct Tscannedonly *)handle->
				data)->activefiles[fsp->fh->fd]);
		flush_sendbuffer(handle);
		SAFE_FREE(((struct Tscannedonly *)handle->
			   data)->activefiles[fsp->fh->fd]);
		((struct Tscannedonly *)handle->data)->activefiles[fsp->fh->
								   fd] = NULL;
	}
	SMB_VFS_NEXT_CLOSE(handle, fsp);
}
#endif

static int scannedonly_rename(vfs_handle_struct * handle,
			      const char *oldname, const char *newname)
{
	/* rename the cache file too */
	char *cachefile_old, *cachefile_new;

	cachefile_old = cachefile_name_f_fullpath(oldname);
	cachefile_new = cachefile_name_f_fullpath(newname);
	if (SMB_VFS_NEXT_RENAME(handle, cachefile_old, cachefile_new) != 0) {
		DEBUG(5,
		      ("failed to rename %s into %s\n", cachefile_old,
		       cachefile_new));
	}
	SAFE_FREE(cachefile_old);
	SAFE_FREE(cachefile_new);
	return SMB_VFS_NEXT_RENAME(handle, oldname, newname);
}

static int scannedonly_unlink(vfs_handle_struct * handle, const char *path)
{
	/* unlink the 'scanned' file too */
	char *cachefile = cachefile_name_f_fullpath(path);
	if (SMB_VFS_NEXT_UNLINK(handle, cachefile) != 0) {
		DEBUG(5, ("failed to unlink %s\n", cachefile));
	}
	SAFE_FREE(cachefile);
	return SMB_VFS_NEXT_UNLINK(handle, path);
}

static int scannedonly_rmdir(vfs_handle_struct * handle, const char *path)
{
	/* if there are only .scanned: .virus: or .failed: files, we delete 
	those, because the client cannot see them */
	DIR *dirp;
	SMB_STRUCT_DIRENT *dire;
	int only_deletable_files = 1, have_files = 0;

	if (!((struct Tscannedonly *)handle->data)->rm_hidden_files_on_rmdir)
		return SMB_VFS_NEXT_RMDIR(handle, path);

	char *path_w_slash = name_w_ending_slash(path);
	dirp = SMB_VFS_NEXT_OPENDIR(handle, path, NULL, 0);
	while ((dire = MY_NEXT_READDIR(handle, dirp)) != NULL) {
		if (strcmp(dire->d_name, ".") == 0
		    || strcmp(dire->d_name, "..") == 0) {
			continue;
		}
		have_files = 1;
		if (!is_scannedonly_file(dire->d_name)) {
			if (((struct Tscannedonly *)handle->
			     data)->show_special_files == 0) {
				SMB_STRUCT_STAT sbuf1;
				char *fullpath;
				int retval;
				/* stat the file and see if it is a 
				special file */
				fullpath =
				    path_plus_name(path_w_slash, dire->d_name);
				retval =
				    SMB_VFS_NEXT_STAT(handle, fullpath, &sbuf1);
				if (retval == 0 && S_ISREG(sbuf1.st_mode)) {
					only_deletable_files = 0;
				}
				SAFE_FREE(fullpath);
			} else {
				only_deletable_files = 0;
			}
			DEBUG(5,
			      ("file found in %s, only_deletable=%d\n",
			       path, only_deletable_files));
			break;
		}
	}
	DEBUG(5,
	      ("path=%s, have_files=%d, only_deletable_files=%d\n",
	       path, have_files, only_deletable_files));
	if (have_files == 1 && only_deletable_files == 1) {
		DEBUG(4,
		      ("scannedonly_rmdir, remove leftover scannedonly "
		       "files from %s\n", path_w_slash));
		SMB_VFS_NEXT_REWINDDIR(handle, dirp);
		while ((dire = MY_NEXT_READDIR(handle, dirp)) != NULL) {
			char *fullpath;
			if (strcmp(dire->d_name, ".") == 0
			    || strcmp(dire->d_name, "..") == 0) {
				continue;
			}
			fullpath = path_plus_name(path_w_slash, dire->d_name);
			DEBUG(5, ("unlink %s\n", fullpath));
			SMB_VFS_NEXT_UNLINK(handle, fullpath);
			SAFE_FREE(fullpath);
		}
	}
	SAFE_FREE(path_w_slash);
	return SMB_VFS_NEXT_CLOSEDIR(handle, dirp);
}

static void free_scannedonly_data(void **data)
{
	SAFE_FREE(*data);
}

static int scannedonly_connect(struct vfs_handle_struct *handle,
			       const char *service, const char *user)
{

	struct Tscannedonly *so;

	so = SMB_MALLOC_P(struct Tscannedonly);
	handle->data = (void *)so;
	handle->free_data = free_scannedonly_data;

	so->domain_socket =
	    lp_parm_bool(SNUM(handle->conn), "scannedonly",
			 "domain_socket", True);
	so->socketname =
	    (char *)lp_parm_const_string(SNUM(handle->conn),
					 "scannedonly", "socketname",
					 "/var/lib/scannedonly/scan");
	so->portnum =
	    lp_parm_int(SNUM(handle->conn), "scannedonly", "portnum", 2020);
	so->scanhost =
	    (char *)lp_parm_const_string(SNUM(handle->conn),
					 "scannedonly", "scanhost",
					 "localhost");

	so->show_special_files =
	    lp_parm_bool(SNUM(handle->conn), "scannedonly",
			 "show_special_files", True);
	so->rm_hidden_files_on_rmdir =
	    lp_parm_bool(SNUM(handle->conn), "scannedonly",
			 "rm_hidden_files_on_rmdir", True);
	so->hide_nonscanned_files =
	    lp_parm_bool(SNUM(handle->conn), "scannedonly",
			 "hide_nonscanned_files", False);
	so->allow_nonscanned_files =
	    lp_parm_bool(SNUM(handle->conn), "scannedonly",
			 "allow_nonscanned_files", False);
	so->scanning_message =
	    (char *)lp_parm_const_string(SNUM(handle->conn),
					 "scannedonly",
					 "scanning_message",
					 "is being scanned for viruses");
	so->scanning_message_len = strlen(so->scanning_message);
	so->recheck_time_open =
	    lp_parm_int(SNUM(handle->conn), "scannedonly",
			"recheck_time_open", 50);
	so->recheck_tries_open =
	    lp_parm_int(SNUM(handle->conn), "scannedonly",
			"recheck_tries_open", 100);
	so->recheck_size_open =
	    lp_parm_int(SNUM(handle->conn), "scannedonly",
			"recheck_size_open", 100);
	so->recheck_time_readdir =
	    lp_parm_int(SNUM(handle->conn), "scannedonly",
			"recheck_time_readdir", 50);
	so->recheck_tries_readdir =
	    lp_parm_int(SNUM(handle->conn), "scannedonly",
			"recheck_tries_readdir", 20);
	memset(so->activefiles, 0, 1024);

	connect_to_scanner(handle);

	return SMB_VFS_NEXT_CONNECT(handle, service, user);
}

/* VFS operations structure */

static vfs_op_tuple scannedonly_op_tuples[] = {
	/* Directory operations */
	{SMB_VFS_OP(scannedonly_opendir), SMB_VFS_OP_OPENDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_readdir), SMB_VFS_OP_READDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_seekdir), SMB_VFS_OP_SEEKDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_telldir), SMB_VFS_OP_TELLDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_rewinddir), SMB_VFS_OP_REWINDDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_closedir), SMB_VFS_OP_CLOSEDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_rmdir), SMB_VFS_OP_RMDIR,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_stat), SMB_VFS_OP_STAT,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_lstat), SMB_VFS_OP_LSTAT,
	 SMB_VFS_LAYER_TRANSPARENT},
	/* File operations */
	{SMB_VFS_OP(scannedonly_open), SMB_VFS_OP_OPEN,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_close), SMB_VFS_OP_CLOSE,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_rename), SMB_VFS_OP_RENAME,
	 SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(scannedonly_unlink), SMB_VFS_OP_UNLINK,
	 SMB_VFS_LAYER_TRANSPARENT},
	/* connection operations */
	{SMB_VFS_OP(scannedonly_connect), SMB_VFS_OP_CONNECT,
	 SMB_VFS_LAYER_TRANSPARENT},
	{NULL, SMB_VFS_OP_NOOP, SMB_VFS_LAYER_NOOP}
};

NTSTATUS vfs_scannedonly_init(void)
{
/*	DEBUG( 9, ("scannedonly vfs_scannedonly_init()\n"));*/
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "scannedonly",
				scannedonly_op_tuples);
}

#ifdef SAMBA_3_0
NTSTATUS init_module(void)
{
/*	DEBUG( 9, ("scannedonly init_module()\n"));*/
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "scannedonly",
				scannedonly_op_tuples);
}
#else
NTSTATUS init_samba_module(void)
{
/*	DEBUG( 9, ("scannedonly init_module()\n"));*/
	return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "scannedonly",
				scannedonly_op_tuples);
}
#endif
